package main

import (
	"flag"
	"fmt"
	"math/big"
	"os"
	"strconv"
	"time"
)

type matrix [2][2]*big.Int

func fastFib(n *big.Int) *big.Int {
	return matrixPower(matrix{{big.NewInt(1), big.NewInt(1)}, {big.NewInt(1), big.NewInt(0)}}, n)[0][1]
}

// m ^ n
func matrixPower(m matrix, n *big.Int) matrix {
	ret := matrix{{big.NewInt(1), big.NewInt(0)}, {big.NewInt(0), big.NewInt(0)}}
	for n.Cmp(big.NewInt(0)) != 0 {
		r := new(big.Int)
		n, r := n.QuoRem(n, big.NewInt(2), r)
		if r.Cmp(big.NewInt(1)) == 0 {
			ret = matrixMultiply(ret, m)
		}

		if n.Cmp(big.NewInt(0)) != 0 {
			m = matrixMultiply(m, m)
		}
	}
	return ret
}

// m1 * m2
func matrixMultiply(m1, m2 matrix) matrix {
	ret := matrix{{big.NewInt(0), big.NewInt(0)}, {big.NewInt(0), big.NewInt(0)}}
	temp := new(big.Int)
	for i := 0; i < 2; i++ {
		for j := 0; j < 2; j++ {
			for k := 0; k < 2; k++ {
				ret[i][j].Add(ret[i][j], temp.Mul(m1[i][k], m2[k][j]))
			}
		}
	}
	return ret
}

var persistFlag bool = false

func main() {
	flag.BoolVar(&persistFlag, "persist", false, "write decimal result to file or not")
	flag.Parse()

	if persistFlag {
		fmt.Printf("%-15s %20s %20s %25s\n", "N", "computing cost (us)", "transformation cost (us)", "binary length")
	} else {
		fmt.Printf("%-15s %20s %25s\n", "N", "computing cost (us)", "binary length")
	}
	for n := 1; n <= 1000000000; n *= 10 {
		var seq []int
		if n <= 1000000 {
			seq = []int{n}
		} else {
			seq = []int{n >> 1, n}
			// seq = []int{ n}
		}
		for _, v := range seq {
			var N *big.Int = big.NewInt(int64(v))
			start := time.Now()
			var ret *big.Int = fastFib(N)
			
			elasped := time.Since(start)
			if persistFlag {
				start := time.Now()
				bs := []byte(ret.String())
				elasped2 := time.Since(start)
				fmt.Printf("%-15d %20d %20d %25d\n", v, elasped.Nanoseconds()/1000, elasped2.Nanoseconds()/1000, ret.BitLen())
				// write it to file
				s := strconv.Itoa(v) + "-result-go.txt"
				f, err := os.OpenFile(s, os.O_WRONLY|os.O_CREATE|os.O_TRUNC, 0666)
				if err != nil {
					fmt.Println("can not open:", s)
				}
				f.Write(bs)
				f.Close()
			} else {
				fmt.Printf("%-15d %20d %25d\n", v, elasped.Nanoseconds()/1000, ret.BitLen())
			}
		}
	}
}
